//****MMM**********MMM**DDDDDDD*****XXX**XXX****//
//****MMMMM******MMMMM**DDD****DDD***XXXXXX*****//
//****MMM*MMM**MMM*MMM**DDD****DDD****XXX******//
//****MMM**MMMMM***MMM**DDD****DDD****XXXX******//
//****MMM**********MMM**DDD****DDD***XXXXXX*****//
//****MMM**********MMM**DDDDDDD*****XXX**XXX****//
//Copyright (C) Michael Matczynski, 1998, 1999
//http://members.lsol.net/michaelm/

// variables - lowercase first letter
// functions - uppercase first letter
// constants - uppercase every letter

// This is the pong demonstration game to show 
// some of the features of MDX, Mike's DirectX Library.

#include "mdx.h"
#include <math.h>  // for trig functions

// left paddle
int lpy = 200;
int lpx = 20;

// right paddle
int rpy = 200;
int rpx = 588;

// ball
int bdir = 110;
int bmoving = 0;
double bspeed = 1.2;
double bx = 288;
double by = 150;

// for ball translations
double cosval[91];
double sinval[91];

//****************************************//
//** P R O T O T Y P E *******************//
//****************************************//
LRESULT CALLBACK WindowProc(HWND hwnd, UINT imsg, WPARAM wp, LPARAM lp);
void FillTrigArrays(void);

//****************************************//
//** W i n d o w P r o c () **************//
//****************************************//

LRESULT CALLBACK WindowProc(HWND hwnd, UINT imsg, WPARAM wp, LPARAM lp)
{
	switch(imsg)
	{	
	case WM_CREATE: 
		{
			//init stuff here
			return 0;
		} break;
	case WM_PAINT:
		{
			PAINTSTRUCT	ps;	// area to paint
			HDC			hdc;	// handle to a device context

			hdc = BeginPaint(hwnd, &ps);	//	start painting
			EndPaint(hwnd, &ps);	//	end painting
			return 0;
		} break;
	case WM_DESTROY:
		{
			PostQuitMessage(0);	//	kill the application
			return 0;
		} break;
	case WM_KEYDOWN:
		{
			switch(wp)
			{
			case VK_ESCAPE:
				{
				SendMessage(hwnd, WM_DESTROY, 0, 0);
				return 0;
				} break;
			}
			return 0;
		} break;
	}	//	end switch

	return (DefWindowProc(hwnd, imsg, wp, lp));	//	process any 
																//	messages we didn't
}	// end WindowProc


//****************************************//
//** W i n M a i n () ********************//
//****************************************//
int WINAPI WinMain(HINSTANCE hinst,
						 HINSTANCE hprevinst,
						 PSTR szcmdline,
						 int icmdshow)
{
	MSG msg;
	WNDCLASSEX wndclass;

	// save the window handle and instance in a global
	g_hinst = hinst;

	wndclass.cbSize			= sizeof(wndclass);
	wndclass.style				= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc		= WindowProc;	//	name of callback function
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hinst;
	wndclass.hIcon				= LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName	= NULL;	//	no menu
	wndclass.lpszClassName	= "dxgame";	//	application class name
	wndclass.hIconSm			= LoadIcon(NULL, IDI_APPLICATION);

	if(!RegisterClassEx(&wndclass))	//	register class
	{
		//error(g_hwnd, 1, "RegisterClassEx()");	// on error
		return(0);
	}	//	end if

	// create the window
	if (!(g_hwnd = CreateWindow("dxgame", // class name
				"DirectX Game",	 // window caption
				WS_POPUP | WS_VISIBLE,	//	window styles
				0,	   // x position 
				0,	   // y position
				GetSystemMetrics(SM_CXSCREEN),	// x size
				GetSystemMetrics(SM_CYSCREEN),	// y size
				NULL,	   // handle to parent 
				NULL,	   // handle to menu
				hinst,// handle to instance
				NULL))) 	// creation parms
	{
		//error(g_hwnd, 2, "CreateWindow()");
		return(0);
	}	//	end if

	FillTrigArrays();

	if(MDXDDInit(640, 480, 16) != MDX_OK)	// directdraw
	{
		return 0;
	}

	if(MDXDSInit() != MDX_OK)	// directsound
	{
		return 0;
	}
	BITMAP_FILE pongbitmap;
	BITMAP_FILE background;
	BITMAP_FILE guy;

	LPDIRECTDRAWSURFACE lpddsleftp;
	LPDIRECTDRAWSURFACE lpddsrightp;
	LPDIRECTDRAWSURFACE lpddsball;
	LPDIRECTDRAWSURFACE lpddsbackground;
	LPDIRECTDRAWSURFACE lpddsguy;

	MDXLoadBitmap(&pongbitmap, "PONG.BMP");
	MDXLoadBitmap(&background, "BG.BMP");
	MDXLoadBitmap(&guy, "MIKE16.BMP");

	lpddsleftp = MDXCreateSurface(32, 176, NULL, _RGB16BIT(31, 31, 0));
	lpddsrightp = MDXCreateSurface(32, 176, NULL, _RGB16BIT(31, 31, 0));
	lpddsball = MDXCreateSurface(64, 64, NULL, _RGB16BIT(31, 31, 0));
	lpddsbackground = MDXCreateSurface(640, 480, NULL, _RGB16BIT(0, 0, 0));
	lpddsguy = MDXCreateSurface(32, 32, NULL, _RGB16BIT(31, 31, 0));

	MDXLoadImageBitmap( lpddsleftp, &pongbitmap, 0, 0, MDX_ABSOLUTE);
	MDXLoadImageBitmap( lpddsrightp, &pongbitmap, 32, 0, MDX_ABSOLUTE);
	MDXLoadImageBitmap( lpddsball, &pongbitmap, 0, 176, MDX_ABSOLUTE);
	MDXLoadImageBitmap( lpddsbackground, &background, 0, 0, MDX_ABSOLUTE);
	MDXLoadImageBitmap( lpddsguy, &guy, 3, 1, MDX_CELL);

	DDBLTFX ddbltfx;
	memset(&ddbltfx,0,sizeof(&ddbltfx));
	ddbltfx.dwSize=sizeof(&ddbltfx);

	MDXUnloadBitmap(&pongbitmap);
	MDXUnloadBitmap(&background);
	MDXUnloadBitmap(&guy);

	int thx_id = MDXLoadWAV("DramaticTheme.WAV");

	if(thx_id == -1)
		return 0;

	int sound_volume = 100;
	MDXPlaySound(thx_id, DSBPLAY_LOOPING);
	

	// enter main event loop
	while(1)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{ 
			// test if this is a quit
			if (msg.message == WM_QUIT)
				break;

			// translate any accelerator keys
			TranslateMessage(&msg);

			// send the message to the window proc
			DispatchMessage(&msg);
		}	// end if

		// main game processing goes here
		//Game_Main();  //temp here


		// left red paddle input
		if(KEY_DOWN(VK_DOWN))
			lpy += 4;
		else if(KEY_DOWN(VK_UP))
			lpy -= 4;
		if(KEY_DOWN(VK_RIGHT))
			lpx += 2;
		else if(KEY_DOWN(VK_LEFT))
			lpx -= 2;

		// left red paddle input
		if(KEY_DOWN(VK_NUMPAD2))
			rpy += 4;
		else if(KEY_DOWN(VK_NUMPAD8))
			rpy -= 4;
		if(KEY_DOWN(VK_NUMPAD6))
			rpx += 2;
		else if(KEY_DOWN(VK_NUMPAD4))
			rpx -= 2;

		// volume control
		if(KEY_DOWN(VK_ADD) && (sound_volume < 100) )
		{
			sound_volume++;
			MDXSetSoundVolume(thx_id, sound_volume);
		}
		else if(KEY_DOWN(VK_SUBTRACT) && (sound_volume > 1) )
		{
			sound_volume--;
			MDXSetSoundVolume(thx_id, sound_volume);
		}

		// should start ball moving?
		if(KEY_DOWN(VK_SPACE))
			bmoving = 1;	// switch the state of ball to moving

		// can change the ball's direction
		if(KEY_DOWN(VK_DIVIDE))
			bdir -= 5;
		if(KEY_DOWN(VK_MULTIPLY))
			bdir += 5;

		// ball translations/bounces
		if(bmoving == 1)
		{
			double xchg = 0;
			double ychg = 0;

			// correct if erraneous solution
			if(bdir < 0)
				bdir += 360;
			else if(bdir > 360)
				bdir -= 360;

			int ldir = bdir;	// get a local var of the ball's dir
			do
			{
				if(ldir > 270) {	// physics algorithm
					ldir -= 270;
					xchg = sinval[ldir] * bspeed;
					ychg = cosval[ldir] * bspeed;

					bx -= ychg;
					by -= xchg;
					break;
				}
				else if(ldir > 180) {
					ldir -= 180;
					xchg = sinval[ldir] * bspeed;
					ychg = cosval[ldir] * bspeed;

					bx -= xchg;
					by += ychg;
					break;
				}
				else if(ldir > 90) {
					ldir -= 90;
					xchg = sinval[ldir] * bspeed;
					ychg = cosval[ldir] * bspeed;

					bx += ychg;
					by += xchg;
					break;
				}
				else {
					xchg = sinval[ldir] * bspeed;
					ychg = cosval[ldir] * bspeed;

					bx += xchg;
					by -= ychg;
					break;
				}
			} while(0);

			// erraneous solution
			if(bdir < 0)
				bdir += 360;
			else if(bdir > 360)
				bdir -= 360;

			// boundary check
			if(bx < 32)
				bdir = 360 - bdir;
			else if(bx > 550)
				bdir = 360 - bdir;

			if(by < 32)
				bdir = 180 - bdir;
			else if(by > 384)
				bdir = 180 - bdir;

		}
		if( (bx < lpx + 32) && (lpy - 32 <= by) && (by <= lpy + 216) )
			bdir = 360 - bdir;

		else if( (bx + 64> rpx) && (rpy - 32<= by) && (by <= rpy + 216) )
			bdir = 360 - bdir;

		// end ball translations

		MDXVsync();
		lpddsback->BltFast(0, 0, lpddsbackground, NULL, DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY );	
		lpddsback->BltFast(lpx, lpy, lpddsleftp, NULL, DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY );
		lpddsback->BltFast(rpx, rpy, lpddsrightp, NULL, DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY );
		lpddsback->BltFast(unsigned long(bx), unsigned long(by), lpddsball, NULL, DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY );

		lpddsback->BltFast(211, 149, lpddsguy, NULL, DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY );
		MDXFlip();

	}	// end while

	// shutdown game and release all resources

	lpddsleftp->Release();
	lpddsrightp->Release();
	lpddsball->Release();
	lpddsbackground->Release();

	MDXDeleteAllSounds();

	MDXDSShutdown();

	MDXShutdown();

	// return to Windows
	return(msg.wParam);
}	// end WinMain


// for ball movement
void FillTrigArrays(void)
{
	FILE * debugfile;

	const double rad2deg = 180 / 3.1415926535;
	const double deg2rad = 3.1415926535 / 180;
	
	debugfile = fopen("trig.txt", "w");
		
		for(int x = 0; x < 91; x++)
	{
		cosval[x] = cos(x * deg2rad);	//	convert rad to deg and store it in cos array
		sinval[x] = sin(x * deg2rad);	//	convert rad to deg and store it in sin array
		fprintf(debugfile, "%f\t%f\n", cosval[x], sinval[x]);

	}

	fclose(debugfile);
}