//****MMM**********MMM**DDDDDDD*****XXX**XXX****//
//****MMMMM******MMMMM**DDD****DDD***XXXXXX*****//
//****MMM*MMM**MMM*MMM**DDD****DDD****XXX******//
//****MMM**MMMMM***MMM**DDD****DDD****XXXX******//
//****MMM**********MMM**DDD****DDD***XXXXXX*****//
//****MMM**********MMM**DDDDDDD*****XXX**XXX****//
//Copyright (C) Michael Matczynski, 1998, 1999
//http://members.lsol.net/michaelm/

// variables - lowercase first letter
// functions - uppercase first letter
// constants - uppercase every letter

#include <windows.h>
#include <ddraw.h>
#include <dsound.h>
#include <io.h>
#include <stdio.h>	// for error functions
#include <fcntl.h>	// for sound defines
#include <mmsystem.h>

#include "mdx.h"

LPDIRECTDRAW lpdd = NULL;	// dd object
LPDIRECTDRAWSURFACE lpddsprimary = NULL;	// screen
LPDIRECTDRAWSURFACE lpddsback = NULL;	// backbuffer
LPDIRECTDRAWCLIPPER lpddclipper = NULL;
HWND g_hwnd = NULL;
HINSTANCE g_hinst = NULL;
int g_screenwidth = 0;
int g_screenheight = 0;
int g_screenbpp = 0;
USHORT *primary_buffer;	// primary video buffer
USHORT *back_buffer;	// secondary back buffer

LPDIRECTSOUND		lpds;           // directsound interface pointer
LPDIRECTSOUNDBUFFER	lpdsbprimary;   // the primary mixing buffer
DSBUFFERDESC		dsbd;           // directsound description
pcm_sound			sound_fx[MAX_SOUNDS];    // the array of secondary sound buffers
WAVEFORMATEX		pcmwf;          // generic waveformat structure

FILE *fp_error;                           // file error functions


///////////////////////////////////////////////////////////////////////
// M D X D D I n i t ( )///////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXDDInit(int width, int height, int bpp)
{
	DDSURFACEDESC ddsd;
	DDSCAPS ddscaps; // a direct draw surface capabilities struct
	
	// get directdraw object
	if(DirectDrawCreate(NULL,&lpdd,NULL)!=DD_OK)
	{
		//error(g_hwnd, 100, "DirectDrawCreate");
		return MDX_ERROR;
	}
	// set cooperation level to windowed mode normal
	if (lpdd->SetCooperativeLevel(g_hwnd,
			DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_ALLOWREBOOT)!=DD_OK)
	{
		//error(g_hwnd, 101, "lpdd->SetCooperativeLevel");
		return MDX_ERROR;
	}
	if (lpdd->SetDisplayMode(width,height,bpp) != DD_OK)
	{
		//error(g_hwnd, 102, "lpdd->SetDisplayMode()");
		return MDX_ERROR;
	}
	g_screenwidth = width;
	g_screenheight = height;
	g_screenbpp = bpp;

	// Create the primary surface
	memset(&ddsd,0,sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;

	// we need to let dd know that we want a complex 
	// flippable surface structure, set flags for that
	ddsd.ddsCaps.dwCaps = 
		DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;

	// set the backbuffer count to 1
	ddsd.dwBackBufferCount = 1;

	// create the primary surface
	if(lpdd->CreateSurface(&ddsd,&lpddsprimary,NULL) != DD_OK)
	{
		return MDX_ERROR;
	}
	// query for the backbuffer i.e the secondary surface
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	
	if(lpddsprimary->GetAttachedSurface(&ddscaps,&lpddsback) != DD_OK)
	{
		return MDX_ERROR;
	}

	// clear out both primary and secondary surfaces
	MDXClearToColor(lpddsprimary, _RGB16BIT(0, 0, 0) );
	MDXClearToColor(lpddsback, _RGB16BIT(0, 0, 0) );

	return MDX_OK;
} // end MDXDDInit

///////////////////////////////////////////////////////////////////////
// M D X S h u t d o w n ( )///////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXShutdown(void)
{
	// this function release all the resources directdraw
	// allocated, mainly to com objects

	// release the clipper first
	if (lpddclipper)
		lpddclipper->Release();

	// release the secondary surface
	if (lpddsback)
		lpddsback->Release();

	// release the primary surface
	if (lpddsprimary)
		lpddsprimary->Release();

	// finally, the main dd object
	if (lpdd)
		lpdd->Release();

	// return success
	return MDX_OK;
} // end MDXShutdown

//BITMAP 
//int MDXCreateBitmap()
//int MDXDestroyBitmap()

//int MDXCreateSurface()
//int MDXDestroySurface()

//int MDXSetClip()
//
//int MDXDrawSprite()
//int MDXStretchSprite()

//int MDXBlit()
//int MDXStretchBlit();

//int MDXClear();

///////////////////////////////////////////////////////////////////////
// M D X C l e a r T o C o l o r ( )///////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXClearToColor(LPDIRECTDRAWSURFACE lpdds, int color)
{
	DDBLTFX ddbltfx; // this contains the DDBLTFX structure

	// clear out the structure and set the size field 
	DD_INIT_STRUCT(ddbltfx);

	// set the dwfillcolor field to the desired color
	ddbltfx.dwFillColor = color; 

	// ready to blt to surface
	lpdds->Blt(NULL,       // ptr to dest rectangle 
		NULL,       // ptr to source surface, NA            
		NULL,       // ptr to source rectangle, NA
		DDBLT_COLORFILL | DDBLT_WAIT,   // fill and wait                   
		&ddbltfx);  // ptr to DDBLTFX structure
	return MDX_OK;
} // end MDXClearToColor


///////////////////////////////////////////////////////////////////////
// M D X V s y n c ( )/////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXVsync(void)
{
	// this function waits for a vertical blank to begin

	lpdd->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);

	// return success
	return MDX_OK;
} // end MDXVsync


///////////////////////////////////////////////////////////////////////
// M D X A t t a c h C l i p p e r ( )/////////////////////////////////
///////////////////////////////////////////////////////////////////////
LPDIRECTDRAWCLIPPER MDXAttachClipper(LPDIRECTDRAWSURFACE lpdds,
                                      int num_rects,
                                      LPRECT clip_list)

{
	// NOT IMPLEMENTED YET
	return (LPDIRECTDRAWCLIPPER) MDX_ERROR;

	// this function creates a clipper from the sent clip list and attaches
	// it to the sent surface

	int index;                         // looping var
	LPDIRECTDRAWCLIPPER lpddclipper;   // pointer to the newly created dd clipper
	LPRGNDATA region_data;             // pointer to the region data that contains
						  // the header and clip list

	// first create the direct draw clipper
	if ((lpdd->CreateClipper(0,&lpddclipper,NULL))!=DD_OK)
		return LPDIRECTDRAWCLIPPER(MDX_ERROR);

	// now create the clip list from the sent data

	// first allocate memory for region data
	region_data = (LPRGNDATA)malloc(sizeof(RGNDATAHEADER)+num_rects*sizeof(RECT));

	// now copy the rects into region data
	memcpy(region_data->Buffer, clip_list, sizeof(RECT)*num_rects);

	// set up fields of header
	region_data->rdh.dwSize          = sizeof(RGNDATAHEADER);
	region_data->rdh.iType           = RDH_RECTANGLES;
	region_data->rdh.nCount          = num_rects;
	region_data->rdh.nRgnSize        = num_rects*sizeof(RECT);

	region_data->rdh.rcBound.left    =  64000;
	region_data->rdh.rcBound.top     =  64000;
	region_data->rdh.rcBound.right   = -64000;
	region_data->rdh.rcBound.bottom  = -64000;

	// find bounds of all clipping regions
	for (index=0; index<num_rects; index++) {
		// test if the next rectangle unioned with the current bound is larger
		if (clip_list[index].left < region_data->rdh.rcBound.left)
			region_data->rdh.rcBound.left = clip_list[index].left;

		if (clip_list[index].right > region_data->rdh.rcBound.right)
			region_data->rdh.rcBound.right = clip_list[index].right;

		if (clip_list[index].top < region_data->rdh.rcBound.top)
			region_data->rdh.rcBound.top = clip_list[index].top;

		if (clip_list[index].bottom > region_data->rdh.rcBound.bottom)
			region_data->rdh.rcBound.bottom = clip_list[index].bottom;

	} // end for index

	// now we have computed the bounding rectangle region and set up the data
	// now let's set the clipping list

	if ((lpddclipper->SetClipList(region_data, 0))!=DD_OK)
	{
		// release memory and return error
		free(region_data);
		return (LPDIRECTDRAWCLIPPER)MDX_ERROR;
	} // end if

	// now attach the clipper to the surface
	if ((lpdds->SetClipper(lpddclipper))!=DD_OK)
	{
		// release memory and return error
		free(region_data);
		return (LPDIRECTDRAWCLIPPER)MDX_ERROR;
	} // end if

	// all is well, so release memory and send back the pointer to the new clipper
	free(region_data);
	return(lpddclipper);

} // end MDXAttachClipper


///////////////////////////////////////////////////////////////////////
// M D X C r e a t e S u r f a c e ( )/////////////////////////////////
///////////////////////////////////////////////////////////////////////
LPDIRECTDRAWSURFACE MDXCreateSurface(int width, int height, int mem_flags, int colorkey)
{
	// this function creates an offscreen plain surface

	DDSURFACEDESC ddsd;         // working description
	LPDIRECTDRAWSURFACE lpdds;  // temporary surface

	// set to access caps, width, and height
	memset(&ddsd,0,sizeof(ddsd));
	ddsd.dwSize  = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;

	// set dimensions of the new bitmap surface
	ddsd.dwWidth  =  width;
	ddsd.dwHeight =  height;

	// set surface to offscreen plain
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | mem_flags;

	// create the surface
	if (lpdd->CreateSurface(&ddsd,&lpdds,NULL)!=DD_OK)
		return (LPDIRECTDRAWSURFACE)MDX_ERROR;

	// set color key to RGB yellow (31, 31, 0)
	DDCOLORKEY color_key; // used to set color key
	color_key.dwColorSpaceLowValue  = colorkey;
	color_key.dwColorSpaceHighValue = colorkey;
	
	// now set the color key for source blitting
	lpdds->SetColorKey(DDCKEY_SRCBLT, &color_key);
	
	// return surface
	return lpdds;
} // end MDXCreateSurface


///////////////////////////////////////////////////////////////////////
// M D X F l i p ( )/////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXFlip(void)
{
	// this function flip the primary surface with the secondary surface

	// test if either of the buffers are locked
	if (primary_buffer || back_buffer)
		return MDX_ERROR;

	// flip pages
	while(lpddsprimary->Flip(NULL, DDFLIP_WAIT)!=DD_OK);

	// flip the surface

	// return success
	return MDX_OK;

} // end MDXFlip


///////////////////////////////////////////////////////////////////////
// M D X L o a d B i t m a p ( )///////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXLoadBitmap(BITMAP_FILE_PTR bitmap, char *filename)
{
	// this function opens a bitmap file and loads the data into bitmap

	int file_handle;  // the file handle
	int index;        // looping index

	UCHAR *temp_buffer = NULL; // used to convert 24 bit images to 16 bit
	OFSTRUCT file_data;        // the file data information

	// open the file if it exists
	if ((file_handle = OpenFile(filename,&file_data,OF_READ))==-1)
	{
		return MDX_ERROR;
	}

	// now load the bitmap file header
	_lread(file_handle, &bitmap->bitmapfileheader,sizeof(BITMAPFILEHEADER));

	// test if this is a bitmap file
	if (bitmap->bitmapfileheader.bfType != 0x4D42)	// universal id for a bitmap
	{
		// close the file
		_lclose(file_handle);

		// return error
		return MDX_ERROR;
	} // end if

	// now we know this is a bitmap, so read in all the sections

	// first the bitmap infoheader

	// now load the bitmap file header
	_lread(file_handle, &bitmap->bitmapinfoheader,sizeof(BITMAPINFOHEADER));

	// now load the color palette if there is one
	if (bitmap->bitmapinfoheader.biBitCount == 8)
	{
		_lread(file_handle, &bitmap->palette,256*sizeof(PALETTEENTRY));

		// now set all the flags in the palette correctly and fix the reversed 
		// BGR RGBQUAD data format
		for (index=0; index < 256; index++)
		{
			// reverse the red and green fields
			int temp_color = bitmap->palette[index].peRed;
			bitmap->palette[index].peRed  = bitmap->palette[index].peBlue;
			bitmap->palette[index].peBlue = temp_color;

			// always set the flags word to this
			bitmap->palette[index].peFlags = PC_NOCOLLAPSE;
		} // end for index

	} // end if

	// finally the image data itself
	_lseek(file_handle,-(int)(bitmap->bitmapinfoheader.biSizeImage),SEEK_END);

	// now read in the image, if the image is 8 or 16 bit then simply read it
	// but if its 24 bit then read it into a temporary area and then convert
	// it to a 16 bit image

	if (bitmap->bitmapinfoheader.biBitCount==8 || bitmap->bitmapinfoheader.biBitCount==16)
	{
		// delete the last image if there was one
		if (bitmap->buffer)
			free(bitmap->buffer);

		// allocate the memory for the image
		if (!(bitmap->buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
		{
			// close the file
			_lclose(file_handle);

			// return error
			return MDX_ERROR;
		} // end if

		// now read it in
		_lread(file_handle,bitmap->buffer,bitmap->bitmapinfoheader.biSizeImage);

	} // end if
	else
	{
		// this must be a 24 bit image, load it in and convert it to 16 bit
		//   printf("\nconverting 24 bit image...");

		// allocate temporary buffer
		if (!(temp_buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
		{
			// close the file
			_lclose(file_handle);

			// return error
			return MDX_ERROR;
		} // end if

		// allocate final 16 bit storage buffer
		if (!(bitmap->buffer=(UCHAR *)malloc(2*bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight)))
		{
			// close the file
			_lclose(file_handle);

			// release working buffer
			free(temp_buffer);

			// return error
			return MDX_ERROR;
		} // end if

		// now read it in
		_lread(file_handle,temp_buffer,bitmap->bitmapinfoheader.biSizeImage);

		// now convert each 24 bit RGB value into a 16 bit value
		for (index=0; index<bitmap->bitmapinfoheader.biWidth*bitmap->bitmapinfoheader.biHeight; index++)
		{
			// extract RGB components (in BGR order), note the scaling
			// to make the 24 bit values into a 16 bit range
			UCHAR blue  = (temp_buffer[index*3 + 0] >> 3),
			green = (temp_buffer[index*3 + 1] >> 3),
			red   = (temp_buffer[index*3 + 2] >> 3);

			// build up 16 bit color word
			USHORT color = _RGB16BIT(red,green,blue);

			// write color to buffer
			((USHORT *)bitmap->buffer)[index] = color;

		} // end for index

		// finally write out the correct number of bits
		bitmap->bitmapinfoheader.biBitCount=16;

	} // end if
	// close the file
	_lclose(file_handle);

	// flip the bitmap
	MDXFlipBitmap(bitmap->buffer, 
	bitmap->bitmapinfoheader.biWidth*(bitmap->bitmapinfoheader.biBitCount/8), 
	bitmap->bitmapinfoheader.biHeight);

	// return success
	return MDX_OK;

} // end MDXLoadBitmap


///////////////////////////////////////////////////////////////////////
// M D X U n L o a d B i t m a p ( )///////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXUnloadBitmap(BITMAP_FILE_PTR bitmap)
{
	// this function releases all memory associated with "bitmap"
	if (bitmap->buffer)
	{
		// release memory
		free(bitmap->buffer);

		// reset pointer
		bitmap->buffer = NULL;

	} // end if

	// return success
	return MDX_OK;

} // end Unload_Bitmap_File


///////////////////////////////////////////////////////////////////////
// M D X F l i p B i t m a p ( )///////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXFlipBitmap(UCHAR *image, int bytes_per_line, int height)
{
	// this function is used to flip upside down .BMP images

	UCHAR *buffer; // used to perform the image processing
	int index;     // looping index

	// allocate the temporary buffer
	if (!(buffer = (UCHAR *)malloc(bytes_per_line*height)))
		return MDX_ERROR;

	// copy image to work area
	memcpy(buffer,image,bytes_per_line*height);

	// flip vertically
	for (index=0; index < height; index++)
			memcpy(&image[((height-1) - index)*bytes_per_line],
			&buffer[index*bytes_per_line], bytes_per_line);

	// release the memory
	free(buffer);

	// return success
	return MDX_OK;

} // end MDXFlipBitmap


///////////////////////////////////////////////////////////////////////
// M D X L o a d I m a g e B i t m a p ( )/////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXLoadImageBitmap(LPDIRECTDRAWSURFACE surface, // bob to load with data
                   BITMAP_FILE_PTR bitmap, // bitmap to scan image data from
                   int cx,int cy,   // cell or absolute pos. to scan image from
                   int mode)        // if 0 then cx,cy is cell position, else 
                                    // cx,cy are absolute coords
{
	// this function extracts a DD surface out of a bitmap file

	UCHAR *source_ptr,   // working pointers for 16bit color
	      *dest_ptr;

	DDSURFACEDESC ddsd;  //  direct draw surface description 

	// set size of the structure
	ddsd.dwSize = sizeof(ddsd);

	// lock the display surface to draw and get ddsd for width&height
	surface->Lock(NULL,
			&ddsd,
			DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR,
			NULL);

	if(mode == MDX_CELL)	// if using cell coords
	{
		//  each square is height/    *** add 1 if the top left cell has 
		//  width + 1 because of *    *   a border on it's outside
		//  added border         *    *
		//                      ***  ***
		cx = cx * (ddsd.dwWidth + 1) + 1;	// calculated absolute coords
		cy = cy * (ddsd.dwHeight + 1) + 1;
	}

	// find location of bitmap data
	source_ptr = bitmap->buffer + 
			( ( (cy*bitmap->bitmapinfoheader.biWidth) + cx) << 1);

	// assign a pointer to the memory surface for manipulation
	dest_ptr = (UCHAR *)ddsd.lpSurface;

	// iterate thru each scanline and copy bitmap
	unsigned int index_y = 0;

	for (index_y=0; index_y < ddsd.dwHeight ; index_y++)
	{
		// copy next line of data to destination
		memcpy(dest_ptr, source_ptr, (ddsd.dwWidth << 1) );

		// advance pointers
		dest_ptr   += (ddsd.lPitch);
		source_ptr += (bitmap->bitmapinfoheader.biWidth << 1);
	} // end for index_y

	// unlock the surface 
	surface->Unlock(ddsd.lpSurface);

	// return success
	return MDX_OK;
} // end MDXLoadImageBitmap


///////////////////////////////////////////////////////////////////////
// M D X O p e n E r r o r F i l e ( )/////////////////////////////
///////////////////////////////////////////////////////////////////////
int MDXOpenErrorFile(char *filename)
{
	// this function opens the output error file

	FILE *fp_temp; // temporary file

	// test if this file is valid
	if ((fp_temp = fopen(filename,"w"))==NULL)
		return MDX_ERROR;

	// close old file if there was one
	if (fp_error)
		MDXCloseErrorFile();

	// assign new file
	fp_error = fp_temp;

	// write out header
	MDXWriteError("\nOpening Error Output File (%s):\n",filename);

	// return success
	return MDX_OK;

} // end MDXOpenErrorFile

///////////////////////////////////////////////////////////

int MDXCloseErrorFile(void)
{
	// this function closes the error file

	if (fp_error)
	{
		// write close file string
		MDXWriteError("\nClosing Error Output File.");

		// close the file handle
		fclose(fp_error);
		fp_error = NULL;

		// return success
		return MDX_OK;
	} // end if
	else
		// error file never existed or another error
		return MDX_ERROR;

} // end Close_Error_File

///////////////////////////////////////////////////////////

int MDXWriteError(char *string, ...)
{
	// this function prints out the error string to the error file

	char buffer[80]; // working buffer

	va_list arglist; // variable argument list

	// make sure both the error file and string are valid
	if (!string || !fp_error)
		return MDX_ERROR;

	// print out the string using the variable number of arguments on stack
	va_start(arglist, string);
	vsprintf(buffer, string, arglist);
	va_end(arglist);

	// write string to file
	fprintf(fp_error, buffer);

	// return success
	return MDX_OK;
} // end MDXWriteError





//******************************************************************************************
//******************************************************************************************
//******************************************************************************************
////NOT YET IMPLEMENTED
//******************************************************************************************
//******************************************************************************************
//******************************************************************************************

int MDXLoadVOC(char *filename)
{
	// this function loads a .voc file, sets up the directsound buffer and loads the data
	// into memory, the function returns the id number of the sound

	int	sound_id = -1,       // id of sound to be loaded
		index,               // looping variable
		data_offset,         // offset to data part of file
		playback_rate,       // playback rate as encoded in file
		data_length;         // length of data

	ULONG  bytesread = 0,    // actual number of bytes read during file read
		   filelength;       // length of file

	int    file_handle;      // general file handle

	UCHAR *snd_buffer,       // temporary sound buffer to hold voc data
		  *audio_ptr_1=NULL, // data ptr to first write buffer 
		  *audio_ptr_2=NULL; // data ptr to second write buffer

	DWORD audio_length_1=0,  // length of first write buffer
	audio_length_2=0;  // length of second write buffer

	// step one: are there any open id's ?
	for (index=0; index < MAX_SOUNDS; index++)
	{	
		// make sure this sound is unused
		if (sound_fx[index].state==SOUND_NULL)
		{
			sound_id = index;
			break;
		} // end if

	} // end for index

	// did we get a free id?
	if (sound_id==-1)
	return MDX_ERROR;

	// step two: load the voc file off disk
	if ((file_handle = _open(filename,_O_BINARY | _O_RDONLY))==-1)
		return MDX_ERROR;

	// get size of file so we can allocate temporary read buffer
	filelength = _filelength(file_handle);

	// allocate a large enough temporary buffer
	snd_buffer = (UCHAR *)malloc(filelength);

	// now read in the data
	bytesread = _read(file_handle, snd_buffer, filelength);

	// access all values and decode VOC encoding for data fields
	data_offset	  = snd_buffer[20];
	playback_rate = (-1000000/(snd_buffer[data_offset+4]-256));
	data_length   = ((*(int *)(snd_buffer+data_offset)) >> 8);

	// set rate and size in data structure
	sound_fx[sound_id].rate  = playback_rate;
	sound_fx[sound_id].size  = data_length;
	sound_fx[sound_id].state = SOUND_LOADED;

	// close the file
	_close(file_handle);

	// step three: create the sound buffer and copy voc data into buffer

	// set up the format data structure
	memset(&pcmwf, 0, sizeof(WAVEFORMATEX));

	pcmwf.wFormatTag	  = WAVE_FORMAT_PCM;  // pulse code modulation
	pcmwf.nChannels		  = 1;                // mono 
	pcmwf.nSamplesPerSec  = 11025;            // always this rate
	pcmwf.nBlockAlign	  = 1;                
	pcmwf.nAvgBytesPerSec = pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
	pcmwf.wBitsPerSample  = 8;
	pcmwf.cbSize		  = 0;


	// prepare to create sounds buffer
	dsbd.dwSize			= sizeof(DSBUFFERDESC);
	dsbd.dwFlags		= DSBCAPS_CTRLDEFAULT | DSBCAPS_STATIC | DSBCAPS_LOCSOFTWARE;
	dsbd.dwBufferBytes	= data_length-NVB_SIZE;
	dsbd.lpwfxFormat	= &pcmwf;

	// create the sound buffer
	if (lpds->CreateSoundBuffer(&dsbd,&sound_fx[sound_id].dsbuffer,NULL)!=DS_OK)
	{
		// release memory
		free(snd_buffer);

		// return error
		return MDX_ERROR;
	} // end if

	// copy data into sound buffer
	if (sound_fx[sound_id].dsbuffer->Lock(0,					 
		  data_length-NVB_SIZE,			
		  (void **) &audio_ptr_1, 
		  &audio_length_1,
		  (void **)&audio_ptr_2, 
		  &audio_length_2,
		  DSBLOCK_FROMWRITECURSOR)!=DS_OK)
		return MDX_ERROR;

	// copy first section of circular buffer
	memcpy(audio_ptr_1, snd_buffer+data_offset+NVB_SIZE, audio_length_1);

	// copy last section of circular buffer
	memcpy(audio_ptr_2, (snd_buffer+data_offset+NVB_SIZE+audio_length_1),audio_length_2);

	// unlock the buffer
	if (sound_fx[sound_id].dsbuffer->Unlock(audio_ptr_1, 
			audio_length_1, 
			audio_ptr_2, 
			audio_length_2)!=DS_OK)
		return MDX_ERROR;

	// release the temp buffer
	free(snd_buffer);

	// return id
	return(sound_id);

} // end MDXLoadVoc

///////////////////////////////////////////////////////////

int MDXLoadWAV(char *filename, int control_flags)
{
	// this function loads a .wav file, sets up the directsound 
	// buffer and loads the data into memory, the function returns 
	// the id number of the sound


	HMMIO 			hwav;    // handle to wave file
	MMCKINFO		parent,  // parent chunk
					child;   // child chunk
	WAVEFORMATEX    wfmtx;   // wave format structure

	int	sound_id = -1,       // id of sound to be loaded
		index;               // looping variable

	UCHAR *snd_buffer,       // temporary sound buffer to hold voc data
		  *audio_ptr_1=NULL, // data ptr to first write buffer 
		  *audio_ptr_2=NULL; // data ptr to second write buffer

	DWORD audio_length_1=0,  // length of first write buffer
		  audio_length_2=0;  // length of second write buffer

	// step one: are there any open id's ?
	for (index=0; index < MAX_SOUNDS; index++)
	{	
		// make sure this sound is unused
		if (sound_fx[index].state==SOUND_NULL)
		{
			sound_id = index;
			break;
		} // end if

	} // end for index

	// did we get a free id?
	if (sound_id==-1)
		return MDX_ERROR;

	// set up chunk info structure
	parent.ckid 	    = (FOURCC)0;
	parent.cksize 	    = 0;
	parent.fccType	    = (FOURCC)0;
	parent.dwDataOffset = 0;
	parent.dwFlags		= 0;

	// copy data
	child = parent;

	// open the WAV file
	if ((hwav = mmioOpen(filename, NULL, MMIO_READ | MMIO_ALLOCBUF))==NULL)
		return MDX_ERROR;

	// descend into the RIFF 
	parent.fccType = mmioFOURCC('W', 'A', 'V', 'E');

	if (mmioDescend(hwav, &parent, NULL, MMIO_FINDRIFF))
	{
		// close the file
		mmioClose(hwav, 0);

		// return error, no wave section
		return MDX_ERROR; 	
	} // end if

	// descend to the WAVEfmt 
	child.ckid = mmioFOURCC('f', 'm', 't', ' ');

	if (mmioDescend(hwav, &child, &parent, 0))
	{
		// close the file
		mmioClose(hwav, 0);

		// return error, no format section
		return MDX_ERROR; 	
	} // end if

	// now read the wave format information from file
	if (mmioRead(hwav, (char *)&wfmtx, sizeof(wfmtx)) != sizeof(wfmtx))
	{
		// close file
		mmioClose(hwav, 0);

		// return error, no wave format data
		return MDX_ERROR;
	} // end if

	// make sure that the data format is PCM
	if (wfmtx.wFormatTag != WAVE_FORMAT_PCM)
	{
		// close the file
		mmioClose(hwav, 0);

		// return error, not the right data format
		return MDX_ERROR; 
	} // end if

	// now ascend up one level, so we can access data chunk
	if (mmioAscend(hwav, &child, 0))
	{
		// close file
		mmioClose(hwav, 0);

		// return error, couldn't ascend
		return MDX_ERROR; 	
	} // end if

	// descend to the data chunk 
	child.ckid = mmioFOURCC('d', 'a', 't', 'a');

	if (mmioDescend(hwav, &child, &parent, MMIO_FINDCHUNK))
	{
		// close file
		mmioClose(hwav, 0);

		// return error, no data
		return MDX_ERROR; 	
	} // end if

	// finally!!!! now all we have to do is read the data in and
	// set up the directsound buffer

	// allocate the memory to load sound data
	snd_buffer = (UCHAR *)malloc(child.cksize);

	// read the wave data 
	mmioRead(hwav, (char *)snd_buffer, child.cksize);

	// close the file
	mmioClose(hwav, 0);

	// set rate and size in data structure
	sound_fx[sound_id].rate  = wfmtx.nSamplesPerSec;
	sound_fx[sound_id].size  = child.cksize;
	sound_fx[sound_id].state = SOUND_LOADED;

	// set up the format data structure
	memset(&pcmwf, 0, sizeof(WAVEFORMATEX));

	pcmwf.wFormatTag	  = WAVE_FORMAT_PCM;  // pulse code modulation
	pcmwf.nChannels		  = 1;                // mono 
	pcmwf.nSamplesPerSec  = 11025;            // always this rate
	pcmwf.nBlockAlign	  = 1;                
	pcmwf.nAvgBytesPerSec = pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
	pcmwf.wBitsPerSample  = 8;
	pcmwf.cbSize		  = 0;

	// prepare to create sounds buffer
	dsbd.dwSize			= sizeof(DSBUFFERDESC);
	dsbd.dwFlags		= control_flags | DSBCAPS_STATIC | DSBCAPS_LOCSOFTWARE;
	dsbd.dwBufferBytes	= child.cksize;
	dsbd.lpwfxFormat	= &pcmwf;

	// create the sound buffer
	if (lpds->CreateSoundBuffer(&dsbd,&sound_fx[sound_id].dsbuffer,NULL)!=DS_OK)
	{
		// release memory
		free(snd_buffer);

		// return error
		return MDX_ERROR;
	} // end if

	// copy data into sound buffer
	if (sound_fx[sound_id].dsbuffer->Lock(0,					 
		  child.cksize,			
		  (void **) &audio_ptr_1, 
		  &audio_length_1,
		  (void **)&audio_ptr_2, 
		  &audio_length_2,
		  DSBLOCK_FROMWRITECURSOR)!=DS_OK)
		return MDX_ERROR;

	// copy first section of circular buffer
	memcpy(audio_ptr_1, snd_buffer, audio_length_1);

	// copy last section of circular buffer
	memcpy(audio_ptr_2, (snd_buffer+audio_length_1),audio_length_2);

	// unlock the buffer
	if (sound_fx[sound_id].dsbuffer->Unlock(audio_ptr_1, 
			audio_length_1, 
			audio_ptr_2, 
			audio_length_2)!=DS_OK)
		return MDX_ERROR;

	// release the temp buffer
	free(snd_buffer);

	// return id
	return(sound_id);

} // end MDXLoadWav

///////////////////////////////////////////////////////////

int MDXReplicateSound(int source_id)
{
	// this function replicates the sent sound and sends back the
	// id of the replicated sound, you would use this function
	// to make multiple copies of a gunshot or something that
	// you want to play multiple times simulataneously, but you
	// only want to load once

	if (source_id!=-1)
	{
		// duplicate the sound buffer
		// first hunt for an open id

		for (int id=0; id < MAX_SOUNDS; id++)
		{
			// is this sound open?
			if (sound_fx[id].state==SOUND_NULL)
			{
				// first make an identical copy
				sound_fx[id] = sound_fx[source_id];

				// now actually replicate the directsound buffer
				if (lpds->DuplicateSoundBuffer(sound_fx[source_id].dsbuffer,
									   &sound_fx[id].dsbuffer)!=DS_OK)
				{
					// reset sound to NULL
					sound_fx[id].dsbuffer = NULL;
					sound_fx[id].state    = SOUND_NULL;

					// return error
					 return MDX_ERROR;
				} // end if

				// now fix up id
				sound_fx[id].id = id;

				// return replicated sound
				return(id);

			} // end if found

		} // end for id

	} // end if
	else
		return MDX_ERROR;

	// else failure
	return MDX_ERROR;

} // end MDXReplicateSound







///////////////////////////////////////////////////////////

int MDXDSInit(void)
{
	// this function initializes the sound system
	static int first_time = 1; // used to track the first time the function
							   // is entered

	// test for very first time
	if (first_time)
	{		
		// clear everything out
		memset(sound_fx,0,sizeof(pcm_sound)*MAX_SOUNDS);

		// reset first time
		first_time = 0;

		// create a directsound object
		if (DirectSoundCreate(NULL, &lpds, NULL)!=DS_OK )
			return MDX_ERROR;

		// set cooperation level
		if (lpds->SetCooperativeLevel((HWND)g_hwnd,DSSCL_NORMAL)!=DS_OK)
			return MDX_ERROR;

	} // end if

	// initialize the sound fx array
	for (int index=0; index<MAX_SOUNDS; index++)
	{
		// test if this sound has been loaded
		if (sound_fx[index].dsbuffer)
		{
			// stop the sound
			sound_fx[index].dsbuffer->Stop();

			// release the buffer
			sound_fx[index].dsbuffer->Release();

		} // end if

		// clear the record out
		memset(&sound_fx[index],0,sizeof(pcm_sound));

		// now set up the fields
		sound_fx[index].state = SOUND_NULL;
		sound_fx[index].id    = index;

	} // end for index

	// return sucess
	return MDX_OK;

} // end DSInit

///////////////////////////////////////////////////////////

int MDXDSShutdown(void)
{
	// this function releases all the memory allocated and the directsound object
	// itself

	// first turn all sounds off
	MDXStopAllSounds();

	// now release all sound buffers
	for (int index=0; index<MAX_SOUNDS; index++)
		if (sound_fx[index].dsbuffer)
			sound_fx[index].dsbuffer->Release();

	// now release the directsound interface itself
	lpds->Release();

	// return success
	return MDX_OK;

} // end DSShutdown

///////////////////////////////////////////////////////////

int MDXPlaySound(int id, int flags, int volume, int rate, int pan)
{
	// this function plays a sound, the only parameter that 
	// works is the flags which can be 0 to play once or
	// MDXPLAYLOOPING

	if (sound_fx[id].dsbuffer)
	{
		// reset position to start
		if (sound_fx[id].dsbuffer->SetCurrentPosition(0)!=DS_OK)
			return MDX_ERROR;

		// play sound
		if (sound_fx[id].dsbuffer->Play(0,0,flags)!=DS_OK)
			return MDX_ERROR;
	} // end if

	// return success
	return MDX_OK;

} // end MDXPlaySound

///////////////////////////////////////////////////////////

int MDXSetSoundVolume(int id,int vol)
{
	// this function sets the volume on a sound 0-100

	if (sound_fx[id].dsbuffer->SetVolume(DSVOLUME_TO_DB(vol))!=DS_OK)
		return MDX_ERROR;

	// return success
	return MDX_OK;

} // end MDXSetSoundVolume

///////////////////////////////////////////////////////////

int MDXSetSoundFreq(int id,int freq)
{
	// this function sets the playback rate

	if (sound_fx[id].dsbuffer->SetFrequency(freq)!=DS_OK)
		return MDX_ERROR;

	// return success
	return MDX_OK;

} // end Set_Sound_Freq

///////////////////////////////////////////////////////////

int MDXSetSoundPan(int id,int pan)
{
	// this function sets the pan, -10,000 to 10,0000

	if (sound_fx[id].dsbuffer->SetPan(pan)!=DS_OK)
		return MDX_ERROR;

	// return success
	return MDX_OK;

} // end MDXSetSoundPan

////////////////////////////////////////////////////////////

int MDXStopSound(int id)
{
	// this function stops a sound from playing
	if (sound_fx[id].dsbuffer)
	{
		sound_fx[id].dsbuffer->Stop();
		sound_fx[id].dsbuffer->SetCurrentPosition(0);
	} // end if

	// return success
	return MDX_OK;

} // end MDXStopSound

///////////////////////////////////////////////////////////

int MDXDeleteAllSounds(void)
{
	// this function deletes all the sounds

	for (int index=0; index < MAX_SOUNDS; index++)
		MDXDeleteSound(index);

	// return success always
	return MDX_OK;

} // end MDXDeleteAllSounds

///////////////////////////////////////////////////////////

int MDXDeleteSound(int id)
{
	// this function deletes a single sound and puts it back onto the available list

	// first stop it
	if (!MDXStopSound(id))
		return MDX_ERROR;

	// now delete it
	if (sound_fx[id].dsbuffer)
	{
		// release the com object
		sound_fx[id].dsbuffer->Release();
		sound_fx[id].dsbuffer = NULL;

		// return success
		return MDX_OK;
	} // end if

	// return success if no sound even existed, already done
	return MDX_OK;

} // end MDXDeleteSound

///////////////////////////////////////////////////////////

int MDXStopAllSounds(void)
{
	// this function stops all sounds

	for (int index=0; index<MAX_SOUNDS; index++)
		MDXStopSound(index);	

	// return success
	return MDX_OK;

} // end MDXStopAllSounds

///////////////////////////////////////////////////////////

int MDXStatusSound(int id)
{
	// this function returns the status of a sound
	if (sound_fx[id].dsbuffer)
	{
		ULONG status; 

		// get the status
		sound_fx[id].dsbuffer->GetStatus(&status);

		// return the status
		return(status);

	} // end if
	else // total failure
	return MDX_ERROR;

} // end MDXStatusSound

///////////////////////////////////////////////////////////