#ifndef _MDX1_H_
#define _MDX1_H_


//****MMM**********MMM**DDDDDDD*****XXX**XXX****//
//****MMMMM******MMMMM**DDD****DDD***XXXXXX*****//
//****MMM*MMM**MMM*MMM**DDD****DDD****XXX******//
//****MMM**MMMMM***MMM**DDD****DDD****XXXX******//
//****MMM**********MMM**DDD****DDD***XXXXXX*****//
//****MMM**********MMM**DDDDDDD*****XXX**XXX****//
//Copyright (C) Michael Matczynski, 1998, 1999
//http://members.lsol.net/michaelm/

// variables - lowercase first letter
// functions - uppercase first letter
// constants - uppercase every letter

//****************************************//
//** I N C L U D E ***********************//
//****************************************//
#include <ddraw.h>
#include <dsound.h>
#include <stdio.h>

//****************************************//
//** D E F I N E S ***********************//
//****************************************//
#define MDX_OK 1
#define MDX_ERROR -1

// voc file defines
#define NVB_SIZE       6  // size of new voice block in bytes

#define MAX_SOUNDS     64 // max number of sounds in system at once 

#define SOUND_NULL     0
#define SOUND_LOADED   1
#define SOUND_PLAYING  2
#define SOUND_STOPPED  3

//****************************************//
//** T Y P E S ***************************//
//****************************************//

/*
typedef struct tagBITMAPFILEHEADER
{
	WORD bfType;
	DWORD bfSize;
	WORD bfReserved1;
	WORD bfReserved2;
	DWORD bfOffBits;
} BITMAPFILEHEADER;
*/

// container structure for bitmaps .BMP file
typedef struct BITMAP_FILE_TAG
{
	BITMAPFILEHEADER bitmapfileheader;  // this contains the bitmapfile header
	BITMAPINFOHEADER bitmapinfoheader;  // this is all the info including the palette
	PALETTEENTRY     palette[256];      // we will store the palette here
	UCHAR            *buffer;           // this is a pointer to the data
} BITMAP_FILE, *BITMAP_FILE_PTR;

// this holds a single sound
typedef struct pcm_sound_typ
	{
	LPDIRECTSOUNDBUFFER dsbuffer;   // the ds buffer containing the sound
	int state;                      // state of the sound
	int rate;                       // playback rate
	int size;                       // size of sound
	int id;                         // id number of the sound
	} pcm_sound, *pcm_sound_ptr;


//****************************************//
//** G L O B A L S ***********************//
//****************************************//
extern LPDIRECTDRAW lpdd;	// dd object
extern LPDIRECTDRAWSURFACE lpddsprimary;	// screen
extern LPDIRECTDRAWSURFACE lpddsback;	// backbuffer
extern LPDIRECTDRAWCLIPPER lpddclipper;
extern HWND g_hwnd;
extern HINSTANCE g_hinst;
extern int g_screenwidth;
extern int g_screenheight;
extern int g_screenbpp;
extern USHORT *primary_buffer;	// primary video buffer
extern USHORT *back_buffer;	// secondary back buffer
extern FILE *fp_error;                           // file error functions

// sound
extern LPDIRECTSOUND		lpds;           // directsound interface pointer
extern LPDIRECTSOUNDBUFFER	lpdsbprimary;   // the primary mixing buffer
extern DSBUFFERDESC		dsbd;           // directsound description
extern pcm_sound			sound_fx[MAX_SOUNDS];    // the array of secondary sound buffers
extern WAVEFORMATEX		pcmwf;          // generic waveformat structure

//****************************************//
//** M A C R O S *************************//
//****************************************//
// initializes and sets size of DD structure
#define DD_INIT_STRUCT(ddstruct) { memset(&ddstruct,0,sizeof(ddstruct)); ddstruct.dwSize=sizeof(ddstruct); }
// this builds a 16 bit color value
#define _RGB16BIT(r,g,b) ((b%32) + ((g%32) << 5) + ((r%32) << 10))
//#define _RGB16BIT(r,g,b) (((r/32) << 11) | ((g/32) << 6) | (b/32))	//	kinda works
#define DSVOLUME_TO_DB(volume) ((DWORD)(-30*(100 - volume)))	// for sound

#define MDX_CELL	1
#define MDX_ABSOLUTE	2

// these read the keyboard asynchronously
#define KEY_DOWN(vk_code) ((GetAsyncKeyState(vk_code) & 0x8000) ? 1 : 0)
#define KEY_UP(vk_code)   ((GetAsyncKeyState(vk_code) & 0x8000) ? 0 : 1)
//****************************************//
//** P R O T O T Y P E *******************//
//****************************************//

// general
int MDXDDInit(int width, int height, int bpp);
int MDXShutdown(void);

// directdraw
int MDXClearToColor(LPDIRECTDRAWSURFACE lpdds, int color);
LPDIRECTDRAWSURFACE MDXCreateSurface(int width, int height, int mem_flags, int colorkey);
int MDXVsync(void);
LPDIRECTDRAWCLIPPER MDXAttachClipper(LPDIRECTDRAWSURFACE lpdds,
                                      int num_rects,
                                     LPRECT clip_list);
int MDXLoadBitmap(BITMAP_FILE_PTR bitmap, char *filename);
int MDXUnloadBitmap(BITMAP_FILE_PTR bitmap);
int MDXFlipBitmap(UCHAR *image, int bytes_per_line, int height);

int MDXFlip(void);

LPDIRECTDRAWSURFACE MDXCreateSurface(int width, int height, int mem_flags);

int MDXLoadImageBitmap(LPDIRECTDRAWSURFACE surface, // surface to put data
                   BITMAP_FILE_PTR bitmap, // bitmap to scan image data from
                   int cx,int cy,   // cell or absolute pos. to scan image from
                   int mode);        // if 0 then cx,cy is cell position, else 
                                    // cx,cy are absolute coords

// error file
int MDXOpenErrorFile(char *filename);
int MDXCloseErrorFile(void);
int MDXWriteError(char *string, ...);


// sound
int MDXLoadVOC(char *filename);
int MDXLoadWAV(char *filename, int control_flags = DSBCAPS_CTRLDEFAULT);
int MDXReplicateSound(int source_id);
int MDXPlaySound(int id, int flags=0,int volume=0, int rate=0, int pan=0);
int MDXStopSound(int id);
int MDXStopAllSounds(void);
int MDXDSInit(void);
int MDXDSShutdown(void);
int MDXDeleteSound(int id);
int MDXDeleteAllSounds(void);
int MDXStatusSound(int id);
int MDXSetSoundVolume(int id,int vol);
int MDXSetSoundFreq(int id,int freq);
int MDXSetSoundPan(int id,int pan);


#endif